package main

import (
	"archive/zip"
	_ "embed"
	"fmt"
	"io"
	"os"
	"os/exec"
	"path"
	"path/filepath"
	"regexp"
	"runtime"
	"strconv"
	"strings"
	"text/template"
	"time"

	"github.com/evanw/esbuild/pkg/api"
	_ "github.com/go-sql-driver/mysql"
	_ "github.com/lib/pq"
	"github.com/ssgo/dao/dao"
	"github.com/ssgo/db"
	"github.com/ssgo/httpclient"
	"github.com/ssgo/tool/sskey/sskeylib"
	"github.com/ssgo/u"
	"github.com/tdewolff/minify/v2"
	"github.com/tdewolff/minify/v2/html"
	_ "modernc.org/sqlite"
)

var version = "v0.0.6"

//go:embed templates/_main.go
var mainCodeTPL string

//go:embed templates/_cacheFiles.go
var cacheFilesTPL string

//go:embed templates/_gitignore
var gitignoreTPL string

//go:embed templates/_Dao.js
var daoTPL string

//go:embed templates/_autoDao.js
var autoDaoTPL string

//go:embed templates/_daoIndex.js
var daoIndexTPL string

//go:embed templates/_sample.js
var daoSampleTPL string

var goPath = "go"

type Command struct {
	Name      string
	ShortName string
	Args      string
	Comment   string
	Func      func([]string) bool
}

type Config struct {
	Id            string
	Name          string
	Version       string
	Main          string
	Target        map[string]string
	Module        map[string]string
	ModuleAlias   map[string]string
	ExtraImport   map[string]string
	CacheFile     []string
	SSKey         string
	CgoEnable     bool
	BuildFrom     string
	BuildEnv      map[string][]string
	BuildLdFlags  map[string]string
	WatchPath     string
	WatchFileType string
	BuildWeb      struct {
		From      string
		To        string
		CopyFiles map[string]string
	}
	WatchWeb string
}

var commands = []Command{
	{"init", "i", "[service|client]", "init a new project for empty dir", initProject},
	{"build", "b", "[all|os arch]", "build project", buildProject},
	{"build web", "bw", "[all|os arch]", "build project", buildWeb},
	{"pack", "p", "[all|os arch]", "pack project", packProject},
	{"pack web", "pw", "[all|os arch]", "pack project", packWeb},
	// {"login", "l", "", "login to apigo.cloud", login},
	// {"deploy", "dp", "", "deploy project to apigo.cloud", depolyProject},
	{"run", "r", "[args...]", "run project", runProject},
	{"run build", "rb", "[args...]", "build and run project", buildRunProject},
	{"run watch build", "rwb", "[args...]", "build and run project", buildRunWatchProject},
	{"run watch", "rw", "[args...]", "watch and run project", watchRunProject},
	{"test", "t", "[args...]", "test project", testProject},
	{"test build", "tb", "[args...]", "build and test project", buildTestProject},
	{"test watch", "tw", "[args...]", "watch and test project", watchTestProject},
	{"tidy", "td", "", "tidy project, run go mod tidy", tidyProject},
	{"check", "ck", "", "check go modules version", checkProject},
	{"export", "e", "", "export typescript code for dev", exportForDev},
	{"make db", "db", "", "make db from desc file", makeDB},
	{"make dao", "dao", "", "make dao from desc file", makeDao},
	{"make er", "er", "", "make er file from desc file", makeER},
	{"tags", "", "", "show git tags", showGitTags},
	{"add tag", "t+", "[version]", "add git tag and push, if no new tag specified will use last tag +1", addGitTag},
	{"commit", "c+", "comment", "commit git repo and push, comment is need", commitGitRepo},
	{"set proxy", "sp", "[proxy_url]", "set go proxy, default is https://goproxy.cn,direct", setGoProxy},
	{"set npm", "sn", "[mirror_url]", "set npm mirror, default is https://registry.npm.taobao.org/", setNpmMirror},
}

var hc = httpclient.GetClient(30 * time.Second)
var cachePath = ".ag"

func init() {
	hc.SetGlobalHeader("Content-Type", "application/json")
	if homePath, err := os.UserHomeDir(); err == nil {
		cachePath = filepath.Join(homePath, ".ag", "cache")
	}
}

func setGoProxy(args []string) bool {
	mirrorUrl := "https://goproxy.cn,direct"
	if len(args) > 0 {
		mirrorUrl = args[0]
	}
	return nil == runCommand("go", "env", "-w", "GO111MODULE=on") && nil == runCommand("go", "env", "-w", "GOPROXY="+mirrorUrl)
}

func setNpmMirror(args []string) bool {
	mirrorUrl := "https://registry.npm.taobao.org/"
	if len(args) > 0 {
		mirrorUrl = args[0]
	}
	return nil == runCommand("npm", "config", "set", "registry", mirrorUrl)
}

func showGitTags(args []string) bool {
	return nil == runCommand("git", "tag", "-l", "v*", "--sort=-taggerdate", "--format=%(refname:short) %(taggerdate:short) %(*objectname:short)")
}

func commitAndTagGitRepo(args []string) bool {
	if commitGitRepo(args) {
		return addGitTag([]string{})
	}
	return false
}

func commitGitRepo(args []string) bool {
	comment := strings.Join(args, " ")
	if comment != "" {
		if err := runCommand("git", "commit", "-a", "-m", comment); err == nil {
			if err := runCommand("git", "push"); err == nil {
				return true
			} else {
				fmt.Println("git push failed:", err.Error())
			}
		} else {
			fmt.Println("git commit failed:", err.Error())
		}
	} else {
		fmt.Println("commit message is empty")
	}
	return false
}

func addGitTag(args []string) bool {
	newVer := ""
	if len(args) > 0 {
		newVer = args[0]
	}
	if newVer == "" {
		if outs, err := u.RunCommand("git", "tag", "-l", "v*", "--sort=taggerdate"); err == nil {
			oldVer := "v0.0.0"
			for i := len(outs) - 1; i >= 0; i-- {
				if outs[i][0] == 'v' && strings.IndexByte(outs[i], '.') != -1 {
					oldVer = outs[len(outs)-1]
					break
				}
			}

			if len(args) > 0 {
				newVer = args[0]
			} else {
				versionParts := strings.Split(oldVer, ".")
				v, err := strconv.Atoi(versionParts[len(versionParts)-1])
				if err != nil {
					v = 0
				}
				versionParts[len(versionParts)-1] = strconv.Itoa(v + 1)
				newVer = strings.Join(versionParts, ".")
			}
		} else {
			fmt.Println("get last tag failed:", err.Error())
		}
	}

	if newVer != "" {
		if err := runCommand("git", "tag", "-a", newVer, "-m", "update tag by 'ag tag+'"); err == nil {
			if err := runCommand("git", "push", "origin", newVer); err != nil {
				fmt.Println("git push failed:", err.Error())
			}
		} else {
			fmt.Println("git add tag failed:", err.Error())
		}
	}
	return false
}

func findTool() (sskeyPath string, logVPath string) {
	sskeyPath = "sskey"
	logVPath = "logv"
	if binPath, err := exec.LookPath("sskey"); err == nil && binPath != "" {
		sskeyPath = binPath
	}
	if binPath, err := exec.LookPath("logv"); err == nil && binPath != "" {
		logVPath = binPath
	}

	if sskeyPath == "sskey" || logVPath == "logv" {
		_ = runCommand(goPath, "get", "-u", "github.com/ssgo/tool")
		if sskeyPath == "sskey" {
			_ = runCommand(goPath, "install", "github.com/ssgo/tool/sskey")
		}
		if logVPath == "logv" {
			_ = runCommand(goPath, "install", "github.com/ssgo/tool/logv")
		}
	}
	return sskeyPath, logVPath
}

func parseRepo(url string) (apiUrl, owner, repo string) {
	name := ""
	if strings.HasPrefix(url, "github.com/") {
		apiUrl = "https://api.github.com/"
		name = url[11:]
	} else if strings.HasPrefix(url, "gitee.com/") {
		apiUrl = "https://gitee.com/api/v5/"
		name = url[10:]
	} else if strings.HasPrefix(url, "apigo.cc/") {
		apiUrl = "https://apigo.cc/api/v1/"
		name = url[16:]
	} else {
		apiUrl = "https://apigo.cc/api/v1/"
		if strings.ContainsRune(url, '/') {
			name = url
		} else {
			name = "ag/" + url
		}
	}

	if !strings.ContainsRune(name, '/') {
		name = name + "/" + name
	}

	a := strings.Split(name, "/")
	owner = a[0]
	repo = a[1]

	return
}

type Tag struct {
	Name   string
	Commit struct {
		Sha string
		Url string
	}
	Zipball_url string
	Tarball_url string
}

func fetchRepo(name string) (repoPath string) {
	apiUrl, owner, repo := parseRepo(name)
	tagsUrl := apiUrl + filepath.Join("repos", owner, repo, "tags")
	tags := make([]Tag, 0)
	fmt.Println("try to fetch tags", tagsUrl)
	r := hc.Get(tagsUrl)
	if r.Error == nil {
		r.To(&tags)
	} else if apiUrl != "https://api.github.com/" {
		tagsUrl = "https://api.github.com/" + filepath.Join("repos", owner, repo, "tags")
		fmt.Println("try to fetch tags", tagsUrl)
		r = hc.Get(tagsUrl)
		if r.Error == nil {
			r.To(&tags)
		} else {
			fmt.Println("fetch tags error", r.Error)
		}
	}
	if len(tags) > 0 {
		lastVersion := tags[0].Name
		zipUrl := tags[0].Zipball_url
		tagPath := filepath.Join(cachePath, owner, repo, lastVersion)
		if !u.FileExists(tagPath) {
			zipFile := filepath.Join(cachePath, owner, repo, lastVersion+".zip")
			fmt.Println("Download file ", zipUrl)
			_, err := hc.Download(zipFile, zipUrl, func(start, end int64, ok bool, finished, total int64) {
				fmt.Printf(" %.2f%%", float32(finished)/float32(total)*100)
			})
			fmt.Println()
			if err != nil {
				fmt.Println(u.BRed("download file failed"))
				return
			}
			if u.FileExists(zipFile) {
				if reader, err := zip.OpenReader(zipFile); err == nil {
					for _, f := range reader.File {
						toFile := filepath.Join(tagPath, f.Name)
						if f.FileInfo().IsDir() {
							//fmt.Println("extract dir", f.Name, toFile)
							os.MkdirAll(toFile, 0755)
						} else {
							//fmt.Println("extract file", f.Name, toFile)
							u.CheckPath(toFile)
							if fp, err := os.OpenFile(toFile, os.O_RDWR|os.O_CREATE|os.O_TRUNC, 0644); err == nil {
								if rd, err := f.Open(); err == nil {
									if _, err := io.Copy(fp, rd); err != nil {
										fmt.Println(u.BRed("write file failed"), toFile, err.Error())
									} else {
										//fmt.Println(u.BGreen("write file success"), toFile)
									}
									_ = rd.Close()
								} else {
									fmt.Println(u.BRed("open file in zip failed"), f.Name, err.Error())
								}
								_ = fp.Close()
							} else {
								fmt.Println(u.BRed("open dst file failed"), toFile, err.Error())
							}
						}
					}
					_ = reader.Close()
				} else {
					fmt.Println(u.BRed("open zip file failed"), zipFile, err.Error())
				}
				_ = os.Remove(zipFile)
			}
		}
		return filepath.Join(tagPath, repo)
	} else {
		fmt.Println(u.BRed("no repo found for"), name)
		return ""
	}
}

func getConfig() *Config {
	conf := &Config{
		Id:          "",
		Name:        "",
		Version:     "v0.0.1",
		Main:        "main.js",
		Module:      map[string]string{},
		ModuleAlias: map[string]string{},
		ExtraImport: map[string]string{},
		CacheFile:   []string{},
		BuildFrom:   "golang",
	}
	u.LoadX("apigo.yml", conf)

	if conf.Name == "" || conf.Id == "" {
		projectPathname := "apigo"
		if pathname, err := os.Getwd(); err == nil {
			projectPathname = filepath.Base(pathname)
		}
		if conf.Name == "" {
			conf.Name = projectPathname
		}
		if conf.Id == "" {
			conf.Id = projectPathname
		}
	}

	if conf.Target == nil {
		conf.Target = map[string]string{
			"darwin":  "amd64 arm64",
			"linux":   "amd64 arm64",
			"windows": "amd64",
		}
	}

	if conf.Module == nil {
		conf.Module = map[string]string{}
	}

	if conf.ModuleAlias == nil {
		conf.ModuleAlias = map[string]string{}
	}

	if conf.ExtraImport == nil {
		conf.ExtraImport = map[string]string{}
	}

	if conf.CacheFile == nil {
		conf.CacheFile = []string{}
	}

	for pkgName, pkgVersion := range conf.Module {
		if pkgVersion == "" {
			conf.Module[pkgName] = "latest"
		}

		aliasName := pkgName
		if strings.HasPrefix(aliasName, "apigo.cc/gojs/") {
			aliasName = aliasName[14:]
		} else if strings.HasPrefix(aliasName, "apigo.cc/") {
			aliasName = aliasName[9:]
		}
		if aliasName != pkgName && conf.ModuleAlias[aliasName] == "" {
			conf.ModuleAlias[aliasName] = pkgName
		}
	}

	for pkgName, pkgVersion := range conf.ExtraImport {
		if pkgVersion == "" {
			conf.ExtraImport[pkgName] = "latest"
		}
	}
	return conf
}

func initProject(args []string) bool {
	// if u.FileExists("main.go") || u.FileExists("go.mod") {
	// 	fmt.Println("main.go or go.mod is exists, are you sure to overwrite it? (y/n)")
	// 	sure := ""
	// 	_, _ = fmt.Scanln(&sure)
	// 	if sure != "y" && sure != "Y" {
	// 		return false
	// 	}
	// }

	if !u.FileExists("apigo.yml") {
		refProjName := "default"
		if len(args) > 0 {
			refProjName = args[0]
		}
		repoPath := fetchRepo(refProjName)
		if u.FileExists(repoPath) {
			for _, f := range u.ReadDirN(repoPath) {
				if !strings.HasPrefix(f.Name, ".") {
					if !u.FileExists(f.Name) {
						if err := u.CopyFile(f.FullName, f.Name); err != nil {
							fmt.Println(u.BRed("copy project file failed"), f.FullName, f.Name, err.Error())
						} else {
							fmt.Println(u.Green("copy project file success"), f.FullName, f.Name)
						}
					}
				}
			}
		} else {
			fmt.Println(u.BRed("fetch repo failed"), repoPath)
			return false
		}
	}

	conf := getConfig()
	writeFile("main.go", mainCodeTPL, conf)

	// modVers := map[string]string{}
	goModStr := ""
	if !u.FileExists("go.mod") {
		_ = runCommand(goPath, "mod", "init", conf.Name)
		_ = runCommand(goPath, "mod", "edit", "-go=1.18")
		findTool()
	} else {
		goModStr = u.ReadFileN("go.mod")
	}

	modCheckList := map[string]string{"apigo.cc/gojs": "latest"}
	for pkgName, pkgVersion := range conf.Module {
		modCheckList[pkgName] = pkgVersion
	}

	for pkgName, pkgVersion := range conf.ExtraImport {
		modCheckList[pkgName] = pkgVersion
	}

	for pkgName, checkVer := range modCheckList {
		if rx, err := regexp.Compile(`(?m)^\s*` + pkgName + `\s+(v[\w-_/.]+)`); err == nil {
			if m := rx.FindStringSubmatch(goModStr); m != nil {
				if m[1] == checkVer || checkVer == "latest" {
					fmt.Println(u.Green("mod " + pkgName + "@" + m[1] + " for " + checkVer + " exists"))
					continue
				}
			}
		}
		_ = runCommand(goPath, "get", pkgName+"@"+checkVer)
	}

	if !u.FileExists(".gitignore") {
		writeFile(".gitignore", gitignoreTPL, nil)
	}

	_ = runCommand(goPath, "mod", "tidy")
	fmt.Println(u.BGreen("new project " + conf.Name + " created"))
	fmt.Println(u.Cyan("run \"ag build\" to use it"))
	fmt.Println(u.Cyan("run \"ag run\" to run main.js"))
	fmt.Println(u.Cyan("run \"ag test\" to run *_test.js"))
	fmt.Println(u.Cyan("run \"ag pack all\" for publish"))
	return true
}

func packProject(args []string) bool {
	return buildProjectWithOption(args, true)
}

func buildProject(args []string) bool {
	return buildProjectWithOption(args, false)
}

func buildWeb(args []string) bool {
	return buildWebWithOption(args, false)
}

func packWeb(args []string) bool {
	return buildWebWithOption(args, true)
}

var cssFileMatcher = regexp.MustCompile(`(?im)^\s*<link.+?text\/css.*?href="(.*?\.css)".*?>`)
var jsFileMatcher = regexp.MustCompile(`(?im)^\s*<script.+?src="(.*?\.js)".*?\/.*?>`)
var cssCodeMatcher = regexp.MustCompile(`(?ims)(^\s*<style>)(.*?)(<\/style>)`)
var jsCodeMatcher = regexp.MustCompile(`(?ims)(^\s*<script>)(.*?)(<\/script>)`)
var viewsMatcher = regexp.MustCompile(`(?i){{Views::(.*?)}}`)

type BuildWebTask struct {
	replace  string
	srcFile  string
	srcCode  string
	result   string
	isOK     bool
	usedTime int64
}

var varsRegexpCache = map[string]*regexp.Regexp{}

func buildWebWithOption(args []string, isPack bool) bool {
	conf := getConfig()
	if conf.BuildWeb.From == "" {
		conf.BuildWeb.From = "src"
	}
	if conf.BuildWeb.To == "" {
		conf.BuildWeb.To = "www"
	}

	// 替换变量
	vars := map[string]string{
		"{{System.Id}}":      conf.Id,
		"{{System.Name}}":    conf.Name,
		"{{System.Version}}": conf.Version,
	}

	os.RemoveAll(conf.BuildWeb.To)
	os.MkdirAll(conf.BuildWeb.To, 0755)
	tmVersion := strconv.FormatInt(time.Now().UnixMilli(), 36)
	filepath.WalkDir(conf.BuildWeb.From, func(srcFilename string, info os.DirEntry, err error) error {
		if err != nil {
			return err
		}
		filename := info.Name()
		if filename[0] == '.' || filename[0] == '_' {
			if info.IsDir() {
				return filepath.SkipDir
			}
			return nil
		}

		if !info.IsDir() {
			ext := filepath.Ext(filename)
			if ext == ".html" || ext == ".htm" {
				taskCount := 0
				taskChan := make(chan bool, 1)
				taskList := make([]*BuildWebTask, 0)
				tt1 := time.Now().UnixMilli()

				// fileInfo, _ := info.Info()
				// lastModified := strconv.FormatInt(fileInfo.ModTime().UnixMilli(), 36)
				// vars["{{LastModified}}"] = lastModified
				vars["{{LastModified}}"] = tmVersion

				// 打包html文件及引用的本地js和css
				fmt.Println(u.BCyan(srcFilename))
				htmlCode := u.ReadFileN(srcFilename)
				htmlPath := filepath.Dir(srcFilename)

				// 打包引用的本地css和js文件
				srcFiles := []string{}
				htmlCode = cssFileMatcher.ReplaceAllStringFunc(htmlCode, func(fullCode string) string {
					if !strings.Contains(fullCode, "//") {
						m := cssFileMatcher.FindStringSubmatch(fullCode)
						srcFiles = append(srcFiles, m[1])
						return strings.Replace(fullCode, m[1], m[1]+"?v="+tmVersion, 1)
					} else {
						return fullCode
					}
				})
				htmlCode = jsFileMatcher.ReplaceAllStringFunc(htmlCode, func(fullCode string) string {
					if !strings.Contains(fullCode, "//") {
						m := jsFileMatcher.FindStringSubmatch(fullCode)
						srcFiles = append(srcFiles, m[1])
						return strings.Replace(fullCode, m[1], m[1]+"?v="+tmVersion, 1)
					} else {
						return fullCode
					}
				})
				// for _, m := range cssFileMatcher.FindAllStringSubmatch(htmlCode, -1) {
				// 	srcFiles = append(srcFiles, m[1])
				// }
				// for _, m := range jsFileMatcher.FindAllStringSubmatch(htmlCode, -1) {
				// 	srcFiles = append(srcFiles, m[1])
				// }
				for _, srcFile := range srcFiles {
					taskCount++
					task := &BuildWebTask{srcFile: srcFile}
					taskList = append(taskList, task)
					go func() {
						task.isOK, task.result, task.usedTime = buildWebFile(true, strings.Contains(task.srcFile, ".css"), task.srcFile, conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack, vars)
						taskChan <- true
					}()
					// ok, msg, usedTime := buildWebFile(true, false, srcFile, conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack)
					// usedTimeStr := fmt.Sprintf("%dms", usedTime)
					// if ok {
					// 	fmt.Println(u.Cyan("  -"), u.Cyan(srcFile), u.BGreen("success"), u.BWhite(usedTimeStr))
					// } else {
					// 	fmt.Println(u.Cyan("  -"), u.Cyan(srcFile), u.BRed("failed"), u.BWhite(usedTimeStr), u.Red(msg))
					// }
				}

				// 打包内嵌的css和js代码
				htmlCode = cssCodeMatcher.ReplaceAllStringFunc(htmlCode, func(fullCode string) string {
					m := cssCodeMatcher.FindStringSubmatch(fullCode)
					taskCount++
					task := &BuildWebTask{replace: "__JS_" + u.UniqueId() + "_" + u.String(taskCount) + "__", srcFile: u.String(taskCount) + ".css", srcCode: fullCode}
					taskList = append(taskList, task)
					go func() {
						task.isOK, task.result, task.usedTime = buildWebFile(false, true, m[2], conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack, vars)
						if task.isOK {
							task.result = m[1] + task.result + m[3]
						}
						taskChan <- true
					}()
					return task.replace
					// ok, msg, usedTime := buildWebFile(false, true, code, conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack)
					// usedTimeStr := fmt.Sprintf("%dms", usedTime)
					// if ok {
					// 	fmt.Println(u.Cyan("  - CSS CODE"), u.BGreen("success"), u.BWhite(usedTimeStr))
					// 	return m[1] + msg + m[3]
					// } else {
					// 	fmt.Println(u.Cyan("  - CSS CODE"), u.BRed("failed"), u.BWhite(usedTimeStr), u.Red(msg))
					// 	return code
					// }
				})
				htmlCode = jsCodeMatcher.ReplaceAllStringFunc(htmlCode, func(fullCode string) string {
					m := jsCodeMatcher.FindStringSubmatch(fullCode)
					taskCount++
					task := &BuildWebTask{replace: "__CSS_" + u.UniqueId() + "_" + u.String(taskCount) + "__", srcFile: u.String(taskCount) + ".js", srcCode: fullCode}
					taskList = append(taskList, task)
					go func() {
						task.isOK, task.result, task.usedTime = buildWebFile(false, false, m[2], conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack, vars)
						if task.isOK {
							task.result = m[1] + task.result + m[3]
						}
						taskChan <- true
					}()
					return task.replace
					// m := jsCodeMatcher.FindStringSubmatch(fullCode)
					// code := m[2]
					// ok, msg, usedTime := buildWebFile(false, false, code, conf.BuildWeb.From, conf.BuildWeb.To, htmlPath, isPack)
					// usedTimeStr := fmt.Sprintf("%dms", usedTime)
					// if ok {
					// 	fmt.Println(u.Cyan("  - JS CODE"), u.BGreen("success"), u.BWhite(usedTimeStr))
					// 	return m[1] + msg + m[3]
					// } else {
					// 	fmt.Println(u.Cyan("  - JS CODE"), u.BRed("failed"), u.BWhite(usedTimeStr), u.Red(msg))
					// 	return code
					// }
				})

				for i := 0; i < taskCount; i++ {
					<-taskChan
				}
				for _, task := range taskList {
					if task.isOK {
						if task.replace != "" {
							htmlCode = strings.Replace(htmlCode, task.replace, task.result, 1)
						}
						fmt.Println(u.Cyan("  -"), u.Cyan(task.srcFile), u.BGreen("success"), u.BWhite(fmt.Sprintf("%dms", task.usedTime)))
					} else {
						if task.replace != "" {
							htmlCode = strings.Replace(htmlCode, task.replace, task.srcCode, 1)
						}
						fmt.Println(u.Cyan("  -"), u.Cyan(task.srcFile), u.BRed("failed"), u.BWhite(fmt.Sprintf("%dms", task.usedTime)), u.Red(task.result))
					}
				}
				fmt.Println("  -", u.BWhite(fmt.Sprintf("%dms", time.Now().UnixMilli()-tt1)))

				// 替换变量(不区分大小写)
				htmlCode = replaceVars(htmlCode, vars)

				// 写入html文件
				dstFilename := strings.Replace(srcFilename, conf.BuildWeb.From, conf.BuildWeb.To, 1)
				if isPack {
					m := minify.New()
					m.AddFunc("text/html", html.Minify)
					var buf strings.Builder
					if err := m.Minify("text/html", &buf, strings.NewReader(htmlCode)); err == nil {
						htmlCode = buf.String()
					}
				}
				u.WriteFile(dstFilename, htmlCode)

				// 生成 index.version
				u.WriteFile(strings.Replace(dstFilename, ext, ".version", 1), tmVersion)
			} else if ext != ".js" && ext != ".css" && ext != ".scss" && ext != ".less" && ext != ".ts" {
				// 非js和css文件复制到www目录
				u.CopyFile(srcFilename, strings.Replace(srcFilename, conf.BuildWeb.From, conf.BuildWeb.To, 1))
			}
		}
		return nil
	})

	// 复制文件
	if len(conf.BuildWeb.CopyFiles) > 0 {
		for src, dst := range conf.BuildWeb.CopyFiles {
			u.CopyFile(src, dst)
		}
	}

	return true
}

func replaceVars(code string, vars map[string]string) string {
	for k, v := range vars {
		rxp := varsRegexpCache[k]
		if rxp == nil {
			rxp = regexp.MustCompile("(?i)" + regexp.QuoteMeta(k))
			varsRegexpCache[k] = rxp
		}
		code = rxp.ReplaceAllString(code, v)
	}
	return code
}

func buildWebFile(isFile, isCSS bool, srcFile, srcRootPath, dstRootPath, htmlPath string, isPack bool, vars map[string]string) (bool, string, int64) {
	var src, dst string
	if isFile {
		src = srcFile
		if strings.HasPrefix(srcFile, "/") {
			src = filepath.Join(srcRootPath, srcFile)
		} else {
			src = filepath.Join(htmlPath, srcFile)
		}
		dst = strings.Replace(src, srcRootPath, dstRootPath, 1)
	}
	useSourceMap := api.SourceMapInline
	if isPack {
		useSourceMap = api.SourceMapNone
	}
	buildOpt := api.BuildOptions{
		Write:             false,
		Bundle:            true,
		Sourcemap:         useSourceMap,
		MinifyWhitespace:  isPack,
		MinifyIdentifiers: isPack,
		MinifySyntax:      isPack,
		Platform:          api.PlatformBrowser,
		// Outdir:            dstRootPath,
		// Loader: map[string]api.Loader{
		// 	".css":   api.LoaderCSS,
		// 	".js":    api.LoaderJS,
		// },
	}
	if isFile {
		// buildOpt.EntryPoints = []string{src}
		srcFile = u.ReadFileN(src)
	}
	if isCSS {
		srcFile = strings.ReplaceAll(srcFile, "url(", "_url(")
	}
	buildOpt.Stdin = &api.StdinOptions{
		Contents:   srcFile,
		ResolveDir: htmlPath,
	}
	if isCSS {
		buildOpt.Stdin.Loader = api.LoaderCSS
	} else {
		buildOpt.Stdin.Loader = api.LoaderJS
	}
	// fmt.Println(u.BMagenta(u.JsonP(buildOpt)))

	t1 := time.Now().UnixMilli()
	result := api.Build(buildOpt)
	t2 := time.Now().UnixMilli() - t1

	code := ""
	if len(result.OutputFiles) > 0 {
		code = string(result.OutputFiles[0].Contents)
	}

	// 替换变量(不区分大小写)
	code = replaceVars(code, vars)

	// 注入视图
	if !isCSS {
		code = viewsMatcher.ReplaceAllStringFunc(code, func(viewsStr string) string {
			m := viewsMatcher.FindStringSubmatch(viewsStr)
			viewPath := m[1]
			if strings.HasPrefix(viewPath, "/") {
				viewPath = filepath.Join(srcRootPath, viewPath)
			} else {
				viewPath = filepath.Join(htmlPath, viewPath)
			}
			if u.FileExists(viewPath) {
				views := map[string]string{}
				filepath.WalkDir(viewPath, func(srcFilename string, info os.DirEntry, err error) error {
					if err != nil {
						return err
					}
					filename := info.Name()
					if filename[0] == '.' {
						if info.IsDir() {
							return filepath.SkipDir
						}
						return nil
					}

					if !info.IsDir() {
						ext := filepath.Ext(filename)
						if ext == ".html" || ext == ".htm" {
							viewName := strings.ReplaceAll(srcFilename[len(viewPath)+1:len(srcFilename)-len(ext)], "/", "-")
							views[viewName] = u.ReadFileN(srcFilename)
						}
					}
					return nil
				})
				return u.Base64String(u.Json(views))
			}
			return viewsStr
		})
	}

	if isCSS {
		code = strings.ReplaceAll(code, "_url(", "url(")
	}

	// 写入目标文件
	if isFile {
		u.WriteFile(dst, code)
	}

	// 返回
	if len(result.Errors) > 0 {
		a := []string{}
		for _, err := range result.Errors {
			a = append(a, err.Text)
		}
		return false, strings.Join(a, "; "), t2
	} else {
		return true, code, t2
	}
}

func buildProjectWithOption(args []string, isPack bool) bool {
	conf := getConfig()
	targets := make([][]string, 0)
	if len(args) > 0 {
		if args[0] == "all" {
			for k, v := range conf.Target {
				for _, arch := range strings.Split(v, " ") {
					targets = append(targets, []string{k, arch})
				}
			}
		} else {
			if len(args) > 1 {
				targets = append(targets, []string{args[0], args[1]})
			} else {
				targets = append(targets, []string{args[0], runtime.GOARCH})
			}
		}
	} else {
		targets = append(targets, []string{runtime.GOOS, runtime.GOARCH})
	}

	output := "build"
	action := "build"
	if isPack {
		output = "release"
		action = "pack"
		cacheFiles := conf.CacheFile
		cachedFiles := make([]string, 0)
		if conf.Main != "" {
			cacheFiles = append([]string{conf.Main}, cacheFiles...)
		}
		for _, file := range cacheFiles {
			if cache := u.LoadFileToB64(file); cache != nil {
				cachedFiles = append(cachedFiles, strings.ReplaceAll(u.Json(cache), `"`, `\"`))
			}
		}
		if len(cachedFiles) > 0 {
			writeFile("cacheFiles.go", cacheFilesTPL, map[string]any{
				"Files": cachedFiles,
			})
		}
		if conf.SSKey != "" {
			homeDir, _ := os.UserHomeDir()
			keyBuf := u.UnBase64(u.ReadFileN(filepath.Join(homeDir, "sskeys", conf.SSKey)))
			if len(keyBuf) == 81 && keyBuf[80] == 217 {
				if sskeyCode, err := sskeylib.MakeCode("go", keyBuf[0:40], keyBuf[40:80]); err == nil {
					u.WriteFile("setSSKey.go", sskeyCode)
				} else {
					fmt.Println(u.Red("failed to make sskey file"), conf.SSKey, err.Error())
				}
			} else {
				fmt.Println(u.Red("no sskey ["+conf.SSKey+"] config found, please run [sskey -c "+conf.SSKey+"] to make"), filepath.Join(homeDir, "sskeys", conf.SSKey))
			}
		}
		defer func() {
			os.Remove("cacheFiles.go")
			os.Remove("setSSKey.go")
		}()
	}

	_ = runCommand(goPath, "mod", "tidy")

	cgoEnable := conf.CgoEnable
	// 未指定时检测是否需要启用CGO
	if !cgoEnable {
		if lines, err := u.RunCommand(goPath, "list", "-f", "{{.ImportPath}}: {{.CgoFiles}}", "all"); err == nil {
			for _, line := range lines {
				line = strings.TrimSpace(line)
				if line == "" || strings.HasSuffix(line, ": []") || strings.HasPrefix(line, "github.com/shirou/gopsutil/") || strings.HasPrefix(line, "net:") {
					continue
				}
				cgoEnable = true
				break
			}
		}
	}

	// 检测是否有docker环境
	upxBin := ""
	_upxArgs := []string{}
	dockerBin := ""
	if dockerBinPath, err := exec.LookPath("docker"); err == nil {
		dockerBin = dockerBinPath
	}

	if isPack {
		if upxBinPath, err := exec.LookPath("upx"); err == nil {
			upxBin = upxBinPath
		} else if dockerBin != "" {
			upxBin = dockerBin
			_upxArgs = append(_upxArgs, "run", "--rm", "-v", ".:/build", "-w", "/build", "apigocc/upx")
		}

		if upxBin == "" {
			if runtime.GOOS == "darwin" {
				fmt.Println(u.Red("upx not found, skip compress"))
				fmt.Println(u.Red("you can install upx by \"brew install upx\""))
			} else if runtime.GOOS == "linux" {
				fmt.Println(u.Red("upx not found, skip compress"))
				fmt.Println(u.Red("you can install upx by \"apt install upx\" or \"yum install upx\" or https://upx.github.io/"))
			} else {
				fmt.Println(u.Red("upx not found, skip compress"))
				fmt.Println(u.Red("you can install upx https://upx.github.io/"))
			}
		}
	}

	for _, target := range targets {
		buildOS := target[0]
		buildArch := target[1]
		name := path.Base(conf.Name)
		ext := ""
		if buildOS == "mac" {
			buildOS = "darwin"
		}
		if buildOS == "windows" {
			ext = ".exe"
		}
		buildPath := filepath.Join(output, fmt.Sprintf("%s_%s_%s%s", name, buildOS, buildArch, ext))

		ldFlags := "-s -w"
		if conf.BuildLdFlags != nil && conf.BuildLdFlags[buildOS] != "" {
			ldFlags = conf.BuildLdFlags[buildOS]
		}

		env := append(conf.BuildEnv[buildOS], "GOOS="+buildOS, "GOARCH="+buildArch)
		buildArgs := []string{}
		buildBin := goPath
		workOS := runtime.GOOS
		useDocker := false
		if cgoEnable {
			env = append(env, "CGO_ENABLED=1")

			if cgoEnable && dockerBin != "" && workOS != buildOS {
				buildBin = dockerBin
				workOS = "linux"
				homePath, _ := os.UserHomeDir()
				buildArgs = append(buildArgs, "run", "--rm", "-v", filepath.Join(homePath, "go")+":/go", "-v", ".:/build", "-w", "/build")
				useDocker = true
			}

			// 检查 CC 工具
			ccBin := ""
			cxxBin := ""
			installCmd := ""
			switch buildOS {
			case "windows":
				ldFlags += " -H windowsgui"
			case "linux":
				// env = append(env, "CGO_LDFLAGS=-static")
			}
			if workOS != buildOS {
				switch workOS + buildOS {
				case "darwinlinux":
					ccBin = "x86_64-linux-musl-gcc"
					cxxBin = "x86_64-linux-musl-g++"
					installCmd = "brew install FiloSottile/musl-cross/musl-cross"
				case "darwinwindows":
					ccBin = "x86_64-w64-mingw32-gcc"
					cxxBin = "x86_64-w64-mingw32-g++"
					installCmd = "brew install mingw-w64"
				case "linuxdarwin":
					ccBin = "o64-clang"
					cxxBin = "o64-clang++"
					installCmd = "apt-get install -y cmake clang bison flex"
				case "linuxwindows":
					ccBin = "x86_64-w64-mingw32-gcc"
					cxxBin = "x86_64-w64-mingw32-g++"
					installCmd = "apt-get install -y gcc-mingw-w64"
				case "windowswindows":
					ccBin = "gcc.exe"
					cxxBin = "g++.exe"
					installCmd = "https://jmeubank.github.io/tdm-gcc/"
				case "windowslinux", "windowsdarwin":
					fmt.Println(u.Red("not support cross compile on windws when CGO_ENABLED=1, you can use WSL"))
					return false
				default:
					fmt.Println(u.Red("not support cross compile on " + buildOS + " " + buildArch + " when CGO_ENABLED=1"))
					return false
				}
			} else {
				if buildOS == "linux" && buildArch == "arm64" {
					ccBin = "aarch64-linux-gnu-gcc"
					cxxBin = "aarch64-linux-gnu-g++"
					installCmd = "apt-get install -y gcc-aarch64-linux-gnu"
				}
			}
			if ccBin != "" && cxxBin != "" {
				if !useDocker {
					ccBinPath, err1 := exec.LookPath(ccBin)
					cxxBinPath, err2 := exec.LookPath(cxxBin)
					if err1 == nil && err2 == nil {
						env = append(env, "CC="+ccBinPath, "CXX="+cxxBinPath)
					} else {
						fmt.Println(u.Red("not found " + ccBin + " or " + cxxBin))
						fmt.Println("you can install by", u.BCyan(installCmd))
						return false
					}
				} else {
					env = append(env, "CC="+ccBin, "CXX="+cxxBin)
				}
			}
		}
		if useDocker {
			for _, v := range env {
				if strings.Contains(v, "%cd%") {
					v = strings.ReplaceAll(v, "%cd%", "/build/")
				}
				buildArgs = append(buildArgs, "-e", v)
			}
			env = nil
			// buildArgs = append(buildArgs, "apigocc/gobuild", "go")
			// buildArgs = append(buildArgs, "gob:v1", "bash", "-c")
			buildArgs = append(buildArgs, conf.BuildFrom, "bash", "-c")
			buildArgs = append(buildArgs, fmt.Sprintln("go", "build", "-o", buildPath, "-ldflags", "'"+ldFlags+"'", "."))
		} else {
			curPath, _ := os.Getwd()
			for i, v := range env {
				if strings.Contains(v, "%cd%") {
					env[i] = strings.ReplaceAll(v, "%cd%", filepath.Join(curPath, ""))
				}
			}
			buildArgs = append(buildArgs, "build", "-o", buildPath, "-ldflags", ldFlags, ".")
		}
		// err := runCommandWithEnv("go", env, "build", "-o", buildPath, "-ldflags", ldFlags, ".")
		err := runCommandWithEnv(buildBin, env, buildArgs...)
		if isPack && upxBin != "" && buildOS != "darwin" {
			upxArgs := append(_upxArgs, buildPath)
			err = runCommand(upxBin, upxArgs...)
			// if upxBin, err := exec.LookPath("upx"); err == nil {
			// 	if buildOS == "darwin" {
			// 		fmt.Println(u.Yellow("upx not support macos, skip compress"))
			// 	} else {
			// 		if err = runCommand(upxBin, buildPath); err == nil {
			// 			fmt.Println(u.Green("compress file success"), buildPath)
			// 		} else {
			// 			fmt.Println(u.BRed("compress file failed"), buildPath, err.Error())
			// 		}
			// 	}
			// } else {
			// 	if workOS == "darwin" {
			// 		fmt.Println(u.Red("upx not found, skip compress"))
			// 		fmt.Println(u.Red("you can install upx by \"brew install upx\""))
			// 	} else if workOS == "linux" {
			// 		fmt.Println(u.Red("upx not found, skip compress"))
			// 		fmt.Println(u.Red("you can install upx by \"apt install upx\" or \"yum install upx\" or https://upx.github.io/"))
			// 	} else {
			// 		fmt.Println(u.Red("upx not found, skip compress"))
			// 		fmt.Println(u.Red("you can install upx https://upx.github.io/"))
			// 	}
			// }
		}
		if err != nil {
			fmt.Println(u.BRed(action+" failed"), buildOS, buildArch, err.Error())
		} else {
			fmt.Println(u.Green(action+" for"), buildOS, buildArch, "to", buildPath)
		}
	}

	if conf.BuildWeb.From != "" {
		buildWebWithOption(args, isPack)
	}

	return true
}

// func CopyFile(from, to string) error {
// 	fromStat, _ := os.Stat(from)
// 	if fromStat.IsDir() {
// 		// copy dir
// 		for _, f := range u.ReadDirN(from) {
// 			err := CopyFile(filepath.Join(from, f.Name), filepath.Join(to, f.Name))
// 			if err != nil {
// 				return err
// 			}
// 		}
// 		return nil
// 	} else {
// 		// copy file
// 		toStat, err := os.Stat(to)
// 		if err == nil && toStat.IsDir() {
// 			to = filepath.Join(to, filepath.Base(from))
// 		}
// 		u.CheckPath(to)
// 		if writer, err := os.OpenFile(to, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0644); err == nil {
// 			defer writer.Close()
// 			if reader, err := os.OpenFile(from, os.O_RDONLY, 0644); err == nil {
// 				defer reader.Close()
// 				_, err = io.Copy(writer, reader)
// 				return err
// 			} else {
// 				return err
// 			}
// 		} else {
// 			return err
// 		}
// 	}
// }

func runProject(args []string) bool {
	_, logVPath := findTool()
	cmdArgs := append([]string{}, args...)
	conf := getConfig()
	name := path.Base(conf.Name)
	binPath := filepath.Join("build", fmt.Sprintf("%s_%s_%s", name, runtime.GOOS, runtime.GOARCH))
	if !u.FileExists(binPath) {
		buildProject([]string{})
	}
	return nil == runCommandPipe(logVPath, binPath, cmdArgs...)
}

func buildRunProject(args []string) bool {
	buildProject(args)
	return runProject(args)
}

func watchRunProject(args []string) bool {
	args = append([]string{"-w"}, args...)
	return runProject(args)
}

func buildRunWatchProject(args []string) bool {
	buildProject(args)
	args = append([]string{"-w"}, args...)
	return runProject(args)
}

func testProject(args []string) bool {
	args = append([]string{"test"}, args...)
	return runProject(args)
}

func buildTestProject(args []string) bool {
	buildProject(args)
	args = append([]string{"test"}, args...)
	return runProject(args)
}

func watchTestProject(args []string) bool {
	args = append([]string{"-w", "test"}, args...)
	return runProject(args)
}

// func watchProject(args []string) bool {
// 	args = append([]string{"-w"}, args...)
// 	return runProject(args)
// }

// func buildWatchProject(args []string) bool {
// 	buildProject(args)
// 	args = append([]string{"-w"}, args...)
// 	return runProject(args)
// }

func exportForDev(args []string) bool {
	return runProject([]string{"-export"})
}

func makeDB(args []string) bool {
	descFile := "DB.txt"
	if len(args) > 0 {
		descFile = args[0]
	}
	if !u.FileExists(descFile) {
		fmt.Println(u.Red("desc file not found" + descFile))
		return false
	}
	erDesc := u.ReadFileN(descFile) + `

_deleted
table v30 PK
id v30 PK
time dt ct
owner v30 I
data t
`

	conn := db.GetDB("default", nil)
	if err := dao.MakeDBFromDesc(conn, erDesc, nil); err == nil {
		fmt.Println(u.Green("make db success"))
		return true
	} else {
		fmt.Println(u.Red("make db failed"), err.Error())
		return false
	}
}

var fieldSizeMatcher = regexp.MustCompile(`\(\d+\)`)

func makeDao(args []string) bool {
	autoDaoTPL = strings.ReplaceAll(autoDaoTPL, "// --- ", "")
	autoDaoTPL = strings.ReplaceAll(autoDaoTPL, "\"--{{", "{{")
	autoDaoTPL = strings.ReplaceAll(autoDaoTPL, "}}--\"", "}}")
	daoIndexTPL = strings.ReplaceAll(daoIndexTPL, "// --- ", "")
	daoIndexTPL = strings.ReplaceAll(daoIndexTPL, "\"--{{", "{{")
	daoIndexTPL = strings.ReplaceAll(daoIndexTPL, "}}--\"", "}}")
	daoSampleTPL = strings.ReplaceAll(daoSampleTPL, "// --- ", "")
	daoSampleTPL = strings.ReplaceAll(daoSampleTPL, "\"--{{", "{{")
	daoSampleTPL = strings.ReplaceAll(daoSampleTPL, "}}--\"", "}}")
	u.WriteFile(filepath.Join("dao", "_Dao.js"), daoTPL)

	descFile := "DB.txt"
	if len(args) > 0 {
		descFile = args[0]
	}
	if !u.FileExists(descFile) {
		fmt.Println(u.Red("desc file not found" + descFile))
		return false
	}
	desc := u.ReadFileN(descFile) + `

_deleted
table v30 PK
id v30 PK
time dt ct
owner v30 I
data t
`

	list := make([]map[string]any, 0)
	for _, g := range dao.MakeERFromDesc(desc) {
		fmt.Println(u.BCyan(g.Name))
		for _, t := range g.Tables {
			pk := make([]string, 0)
			fields := make([]string, 0)
			listFields := make([]string, 0)
			ignoreFields := make([]string, 0)
			ownerField := ""
			versionField := ""
			validField := ""
			addUserField := ""
			addTimeField := ""
			updateUserField := ""
			updateTimeField := ""
			makeID := "null"
			ignore := false
			if strings.Contains(t.Comment, "@ignore") {
				ignore = true
			}
			for _, f := range t.Fields {
				if f.Index == "pk" {
					pk = append(pk, f.Name)
					if len(pk) == 1 {
						if strings.HasPrefix(f.Type, "char") {
							idSize := "12"
							if m := fieldSizeMatcher.FindStringSubmatch(f.Type); m != nil {
								idSize = m[0]
							}
							makeID = "function () { return s.id(this.table, " + idSize + ") }"
						}
					} else {
						makeID = "null"
					}
				}
				if strings.Contains(f.Comment, "@ignore") {
					ignoreFields = append(ignoreFields, f.Name)
				} else {
					listFields = append(listFields, f.Name)
				}
				fields = append(fields, f.Name)
				if strings.Contains(f.Comment, "@version") || (f.Name == "version" && strings.Contains(f.Type, "unsigned") && strings.Contains(f.Type, "int")) {
					versionField = f.Name
				} else if strings.Contains(f.Comment, "@valid") || (f.Name == "isValid" && strings.Contains(f.Type, "tiny") && strings.Contains(f.Type, "int")) {
					validField = f.Name
				} else if strings.Contains(f.Comment, "@owner") || (f.Name == "owner" && (strings.Contains(f.Type, "char") || strings.Contains(f.Type, "int"))) {
					ownerField = f.Name
				} else if strings.Contains(f.Comment, "@addUser") || (f.Name == "addUser" && (strings.Contains(f.Type, "char") || strings.Contains(f.Type, "int"))) {
					addUserField = f.Name
				} else if strings.Contains(f.Comment, "@updateUser") || (f.Name == "updateUser" && (strings.Contains(f.Type, "char") || strings.Contains(f.Type, "int"))) {
					updateUserField = f.Name
				} else if strings.Contains(f.Comment, "@addTime") || (f.Name == "addTime" && f.Type == "datetime") {
					addTimeField = f.Name
				} else if strings.Contains(f.Comment, "@updateTime") || (f.Name == "updateTime" && f.Type == "datetime") {
					updateTimeField = f.Name
				}
			}
			list = append(list, map[string]any{
				"table":           t.Name,
				"pk":              strings.Join(pk, ", "),
				"listFields":      strings.Join(listFields, ", "),
				"searchFields":    strings.Join(listFields, ", "),
				"getFields":       strings.Join(fields, ", "),
				"updateFields":    strings.Join(fields, ", "),
				"ignoreFields":    strings.Join(ignoreFields, ", "),
				"ownerField":      ownerField,
				"versionField":    versionField,
				"validField":      validField,
				"addUserField":    addUserField,
				"addTimeField":    addTimeField,
				"updateUserField": updateUserField,
				"updateTimeField": updateTimeField,
				"makeID":          makeID,
				"ignore":          ignore,
			})
			fmt.Println(u.Cyan("  > "+t.Name), u.BCyan(strings.Join(pk, ", ")), u.BMagenta(validField), u.BMagenta(versionField), u.BMagenta(ownerField), u.BBlue(addUserField), u.BBlue(updateUserField), u.BBlue(addTimeField), u.BBlue(updateTimeField), u.BRed(u.StringIf(ignore, "ignore", "")), u.BYellow(strings.Join(ignoreFields, ", ")))
		}
	}

	if tpl, err := template.New("autoDao").Parse(autoDaoTPL); err == nil {
		if fp, err := os.OpenFile(filepath.Join("dao", "_autoDao.js"), os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0644); err == nil {
			defer fp.Close()
			tpl.Execute(fp, list)
		} else {
			fmt.Println(u.Red(err.Error()))
		}
	} else {
		fmt.Println(u.Red(err.Error()))
	}

	if tpl, err := template.New("daoSample").Parse(daoSampleTPL); err == nil {
		if fp, err := os.OpenFile(filepath.Join("dao", "_sample.js"), os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0644); err == nil {
			defer fp.Close()
			tpl.Execute(fp, list)
		} else {
			fmt.Println(u.Red(err.Error()))
		}
	} else {
		fmt.Println(u.Red(err.Error()))
	}

	daoIndexFile := filepath.Join("dao", "index.js")
	if !u.FileExists(daoIndexFile) {
		if tpl, err := template.New("daoIndex").Parse(daoIndexTPL); err == nil {
			if fp, err := os.OpenFile(daoIndexFile, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0644); err == nil {
				defer fp.Close()
				tpl.Execute(fp, list)
			} else {
				fmt.Println(u.Red(err.Error()))
			}
		} else {
			fmt.Println(u.Red(err.Error()))
		}
	}
	return true
}

func makeER(args []string) bool {
	descFile := "DB.txt"
	erFile := "DB.html"
	for i := 0; i < len(args); i++ {
		if erFile == "DB.html" && strings.HasSuffix(args[i], ".html") {
			erFile = args[i]
		} else if descFile == "DB.txt" {
			descFile = args[i]
		}
	}
	if !u.FileExists(descFile) {
		fmt.Println(u.Red("desc file not found" + descFile))
		return false
	}
	erDesc := u.ReadFileN(descFile)
	dao.MakeERFile(erDesc, "", erFile, nil)
	fmt.Println(u.Green("make er success"))
	return true
}

// func devProject(args []string) bool {
// 	return _runProject(args, true)
// }

// func testProject(args []string) bool {
// 	return _testProject(args, false)
// }

// func devTestProject(args []string) bool {
// 	return _testProject(args, true)
// }

// func _testProject(args []string, isWatch bool) bool {
// 	if u.FileExists("tests") {
// 		if u.FileExists(filepath.Join("tests", "go.mod")) {
// 			_ = os.Chdir("tests")
// 			if isWatch {
// 				args = append(args, "-p", "..")
// 			}
// 			isRun := false
// 			for _, f := range u.ReadDirN(".") {
// 				if strings.HasSuffix(f.Name, ".go") {
// 					goStr := u.ReadFileN(f.FullName)
// 					if strings.Contains(goStr, "package main") && !strings.Contains(goStr, "package main_test") {
// 						isRun = true
// 					}
// 					break
// 				}
// 			}
// 			if !isRun {
// 				args = append(args, "test", "-v", ".")
// 			} else {
// 				args = append(args, "run", ".")
// 			}
// 		} else {
// 			args = append(args, "test", "-v", "tests")
// 		}
// 	} else {
// 		args = append(args, "test", "-v", ".")
// 	}
// 	_ = runCommand(goPath, "mod", "tidy")
// 	sskeyPath, logVPath := findTool()
// 	if isWatch {
// 		args2 := append([]string{"-pt", ".go,.js,.yml"}, args...)
// 		_ = runCommandPipeWithEnv(logVPath, sskeyPath, goRunEnv, args2...)
// 	} else {
// 		_ = runCommandPipeWithEnv(logVPath, goPath, goRunEnv, args...)
// 	}
// 	return true
// }

// var pkgMatcher = regexp.MustCompile(`(?im)^\s*(import)?\s*_\s+"([\w-_/.]+)"`)
// var importMatcher = regexp.MustCompile(`(?im)^\s*import\s+([\w{}, ]+)\s+from\s+['"]([\w./\\\- ]+)['"]`)

// func makeJsImports(path string, jsImports *[]string, pkgName string) {
// 	if u.FileExists(filepath.Join(path, "go.mod")) {
// 		if m := modNameMatcher.FindStringSubmatch(u.ReadFileN(filepath.Join(path, "go.mod"))); m != nil {
// 			if pkgName != m[1] {
// 				return
// 			}
// 		}
// 	}
// 	for _, f := range u.ReadDirN(path) {
// 		filename := f.Name
// 		fullName := filepath.Join(path, filename)
// 		if !strings.HasPrefix(filename, ".") && !strings.HasPrefix(filename, "_") && filename != "_makePluginCode" && filename != "node_modules" && filename != "www" && filename != "src" {
// 			if f.IsDir {
// 				makeJsImports(fullName, jsImports, pkgName)
// 			} else if strings.HasSuffix(filename, ".js") {
// 				if code, err := u.ReadFile(fullName); err == nil {
// 					for _, m := range importMatcher.FindAllStringSubmatch(code, 1024) {
// 						if !strings.HasPrefix(m[2], ".") && !strings.HasPrefix(m[2], "_") && m[2] != "current-plugin" && m[2] != "console" && m[2] != "logger" {
// 							checkFile := filepath.Join(path, m[2])
// 							checkFileInfo := u.GetFileInfo(checkFile)
// 							if checkFileInfo != nil && checkFileInfo.IsDir {
// 								checkFile = filepath.Join(checkFile, "index.js")
// 							} else if !strings.HasSuffix(checkFile, ".js") {
// 								checkFile += ".js"
// 							}
// 							//fmt.Println(u.BMagenta(checkFile), u.FileExists(checkFile))
// 							if !u.FileExists(checkFile) {
// 								*jsImports = u.AppendUniqueString(*jsImports, m[2])
// 							}
// 						}
// 					}
// 				}
// 			}
// 		}
// 	}
// }

// func makePluginCodeImports(from string, imports *[]string, parentModuleName string) {
// 	jsImports := make([]string, 0)
// 	currentParentModuleName := parentModuleName
// 	if u.FileExists(filepath.Join(from, "go.mod")) {
// 		pkgName := ""
// 		if m := modNameMatcher.FindStringSubmatch(u.ReadFileN(filepath.Join(from, "go.mod"))); m != nil {
// 			pkgName = m[1]
// 		}
// 		makeJsImports(from, &jsImports, pkgName)
// 		parentModuleName = pkgName
// 	}

// 	currentPkgName := ""
// 	for _, f := range u.ReadDirN(from) {
// 		if f.IsDir {
// 			if !strings.HasPrefix(f.Name, ".") {
// 				makePluginCodeImports(filepath.Join(from, f.Name), imports, parentModuleName)
// 			}
// 		} else {
// 			if strings.HasSuffix(f.Name, ".go") && !strings.HasPrefix(f.Name, ".") && f.Name != "makePluginCode.go" {
// 				if code, err := u.ReadFile(filepath.Join(from, f.Name)); err == nil {
// 					if m := pkgNameMatcher.FindStringSubmatch(code); m != nil {
// 						if currentPkgName+"_test" != m[1] {
// 							currentPkgName = m[1]
// 						}
// 					}
// 					for _, m := range pkgMatcher.FindAllStringSubmatch(code, 1024) {
// 						if m[2] != "current-plugin" && !strings.HasPrefix(f.Name, "jsImports") {
// 							*imports = u.AppendUniqueString(*imports, m[2])
// 						}
// 					}
// 				}
// 			}
// 		}
// 	}

// 	if currentPkgName != "" && u.FileExists(filepath.Join(from, "go.mod")) {
// 		pkgList := make([]string, 0)
// 		for _, plgPkg := range jsImports {
// 			if plgPkg != currentParentModuleName && !u.StringIn(*imports, plgPkg) {
// 				pkgList = append(pkgList, plgPkg)
// 				*imports = u.AppendUniqueString(*imports, plgPkg)
// 			}
// 		}
// 		if len(pkgList) > 0 {
// 			_ = u.WriteFile(filepath.Join(from, u.StringIf(strings.HasSuffix(currentPkgName, "_test"), "jsImports_test.go", "jsImports.go")), `package `+currentPkgName+`
// import _ "`+strings.Join(pkgList, "\"\nimport _ \"")+`"
// `)
// 		}
// 	}
// }

func writeFile(filename string, fileContent string, data any) {
	if fp, err := os.OpenFile(filename, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0600); err == nil {
		tpl, _ := template.New(filename).Parse(fileContent)
		if err = tpl.Execute(fp, data); err != nil {
			fmt.Println(u.Red(err.Error()))
		}
		_ = fp.Close()
	} else {
		fmt.Println(u.Red(err.Error()))
	}
}

// var replaceMatcher = regexp.MustCompile(`([a-zA-Z0-9._\-/]+)\s+(v[0-9.]+)\s+=>\s+([a-zA-Z0-9._\-/\\]+)`)
// var modNameMatcher = regexp.MustCompile(`(?m)^module\s+([a-zA-Z0-9._\-/]+)$`)
// var pkgNameMatcher = regexp.MustCompile(`(?m)^package\s+([a-zA-Z0-9._\-/]+)$`)

func tidyProject(args []string) bool {
	_ = runCommand(goPath, "mod", "tidy")
	return true
}

func checkProject(args []string) bool {
	if lines, err := u.RunCommand(goPath, "list", "-m", "-u", "all"); err == nil {
		moduleList := map[string]string{}
		indirectModuleList := map[string]string{}
		for _, line := range u.ReadFileLinesN("go.mod") {
			if strings.Contains(line, " v") {
				a := strings.Split(strings.TrimSpace(line), " ")
				if strings.Contains(line, "// indirect") {
					indirectModuleList[a[0]] = a[1]
				} else {
					moduleList[a[0]] = a[1]
				}
			}
		}
		for _, line := range lines {
			a := strings.Split(line, " ")
			if len(a) > 2 {
				if oldVer, ok := moduleList[a[0]]; ok {
					fmt.Println(u.BCyan(a[0]), u.BMagenta(oldVer), u.BGreen(a[2]))
				} else if _, ok := indirectModuleList[a[0]]; ok {
					fmt.Println(u.Cyan(a[0]), u.BMagenta(indirectModuleList[a[0]]), u.BGreen(a[2]))
				} else {
					fmt.Println(a[0], u.BMagenta(a[1]), u.BGreen(a[2]))
				}
			}
		}
	} else {
		fmt.Println(strings.Join(lines, "\n"))
		fmt.Println(u.Red(err.Error()))
	}
	return true
}

func runCommand(name string, args ...string) error {
	return runCommandWithEnv(name, nil, args...)
}

func printCmd(name string, env []string, args []string) {
	fmt.Print(u.BMagenta(name))
	for _, arg := range args {
		fmt.Print(" ", u.BCyan(arg))
	}
	for _, v := range env {
		fmt.Print(" ", u.BYellow(v))
	}
}

func runCommandWithEnv(name string, env []string, args ...string) error {
	// pathname, _ := os.Getwd()
	// fmt.Println(u.BMagenta(pathname), u.BCyan(name), u.Cyan(strings.Join(args, " ")))
	printCmd(name, env, args)
	t1 := time.Now().UnixMilli()
	cmd := exec.Command(name, args...)
	if env != nil {
		// if runtime.GOOS == "windows" {
		// 	env = append(env, "GOTMPDIR="+pathname, "GOWORK="+pathname)
		// }
		cmd.Env = append(os.Environ(), env...)
	}

	cmd.Stdin = os.Stdin
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	err := cmd.Run()
	fmt.Printf(" "+u.BWhite("%dms")+"\n", time.Now().UnixMilli()-t1)
	if err != nil {
		fmt.Println(u.Red(err.Error()))
	}
	return err
}

func runCommandPipe(pipeCommandName, commandName string, args ...string) error {
	return runCommandPipeWithEnv(pipeCommandName, commandName, nil, args...)
}

func runCommandPipeWithEnv(pipeCommandName, commandName string, env []string, args ...string) error {
	printCmd(commandName, env, args)
	t1 := time.Now().UnixMilli()
	cmd1 := exec.Command(commandName, args...)
	cmd2 := exec.Command(pipeCommandName)
	if env != nil {
		// 	if runtime.GOOS == "windows" {
		// 		env = append(env, "GOTMPDIR="+pathname, "GOWORK="+pathname)
		// 	}
		cmd1.Env = append(os.Environ(), env...)
		cmd2.Env = append(os.Environ(), env...)
	}

	r, w := io.Pipe()
	wClosed := false
	defer func() {
		if !wClosed {
			w.Close()
		}
	}()
	cmd1.Stdin = os.Stdin
	cmd1.Stdout = w
	cmd1.Stderr = w
	cmd2.Stdin = r
	cmd2.Stdout = os.Stdout
	cmd2.Stderr = os.Stderr
	var err error
	if err = cmd2.Start(); err == nil {
		if err = cmd1.Start(); err == nil {
			if err = cmd1.Wait(); err == nil {
				w.Close()
				wClosed = true
				// 等待第二个命令完成
				if err = cmd2.Wait(); err == nil {
					return nil
				}
			}
		}
	}
	fmt.Printf(" "+u.BWhite("%dms")+"\n", time.Now().UnixMilli()-t1)
	fmt.Println(u.Red(err.Error()))
	return err
}

func main() {
	var err error
	if goPath, err = exec.LookPath("go"); err != nil || goPath == "" {
		fmt.Println(u.Red("Please install Go SDK first"))
		fmt.Println(u.Cyan("https://go.dev/"))
		return
	}

	if len(os.Args) > 1 {
		cmd1 := os.Args[1]
		cmd2 := cmd1
		if len(os.Args) > 2 {
			cmd2 += " " + os.Args[2]
		}
		for i := len(commands) - 1; i >= 0; i-- {
			cmdInfo := commands[i]
			if len(os.Args) > 2 && (cmd2 == cmdInfo.Name || cmd2 == cmdInfo.ShortName) {
				cmdInfo.Func(os.Args[3:])
				return
			} else if cmd1 == cmdInfo.Name || cmd1 == cmdInfo.ShortName {
				cmdInfo.Func(os.Args[2:])
				return
			}
		}
	}

	fmt.Println("tools for apigo.cc/gojs", version)
	fmt.Println("go sdk", goPath)
	fmt.Println()
	fmt.Println("Usage:")
	fmt.Println(" ", u.Cyan("ag [command] [...]"))
	fmt.Println(" ", u.Magenta("ag [short command] [...]"))
	fmt.Println()
	fmt.Println("Commands:")
	for _, cmdInfo := range commands {
		padStr := ""
		padN := 30 - len(cmdInfo.Name) - len(cmdInfo.ShortName) - len(cmdInfo.Args)
		if padN > 0 {
			padStr = strings.Repeat(" ", padN)
		}
		fmt.Println(" ", u.Cyan(cmdInfo.Name), u.Dim("[")+u.Magenta(cmdInfo.ShortName)+u.Dim("]"), cmdInfo.Args, padStr, cmdInfo.Comment)
	}
	fmt.Println()
	fmt.Println("Examples:")
	fmt.Println(" ", u.Magenta("ag init"), "          init a new default project")
	fmt.Println(" ", u.Magenta("ag init service"), "  init a new service project")
	fmt.Println(" ", u.Magenta("ag init client"), "   init a new client project")
	fmt.Println(" ", u.Magenta("ag init all"), "      init a new project use all modules by apigo.cc/ag included")
	fmt.Println(" ", u.Magenta("ag build"), "         build project")
	fmt.Println(" ", u.Magenta("ag build all"), "     build project for all platforms")
	fmt.Println(" ", u.Magenta("ag run"), "           run project")
	fmt.Println(" ", u.Magenta("ag test"), "          test project")
	fmt.Println()
}
