import u from 'apigo.cc/gojs/util'
import log from 'apigo.cc/gojs/log'
import db from 'apigo.cc/gojs/db'
import redis from 'apigo.cc/gojs/redis'

function makeSelectByFields(dao, isAdmin, fields) {
    if (!fields || fields.length === 0) return ''
    if (isAdmin && (dao.ownerField || dao.validField)) {
        // 管理接口增加管理字段
        fields = u.copy(fields)
        if (dao.ownerField) fields.push(dao.ownerField)
        if (dao.validField) fields.push(dao.validField)
    }
    return '`' + fields.join('`, `') + '`'
}

function fixWhere(where) {
    if (!where) return ''
    if (where.toLowerCase().indexOf('where ') === -1) return 'WHERE ' + where
    return where
}

function makeWhere(dao, isAdmin, query, order, extSql, sqlArgs, owner, version) {
    if (query && Object.keys(query).length > 0) {
        let wheres = []
        sqlArgs = []
        for (let k in query) {
            let v = query[k]
            if (v instanceof Array) {
                wheres.push('`' + k + '` IN ' + db.makeInKeys(v.length))
                sqlArgs.push(...v)
            } else {
                wheres.push('`' + k + '` ' + (v.indexOf('%') === -1 ? '=' : 'LIKE') + ' ?')
                sqlArgs.push(query[k])
            }
        }
        if (dao.versionField && version !== undefined && version !== null) {
            wheres.push('`' + dao.versionField + '` > ?')
            sqlArgs.push(version)
        }
        // 非管理接口增加所属人限制
        if (!isAdmin && dao.ownerField && owner) {
            wheres.push('`' + dao.ownerField + '` = ?')
            sqlArgs.push(owner)
        }
        // 非管理接口增加有效字段限制
        if (!isAdmin && dao.validField) {
            wheres.push('`' + dao.validField + '` = 1')
        }
        extSql = 'WHERE ' + wheres.join(' AND ')
    } else {
        extSql = fixWhere(extSql)
        let appendSql = []
        if (dao.ownerField && owner) extSql = appendSql.push("`" + dao.ownerField + "` = '" + u.string(owner).replace("'", "\\'") + "'")
        if (dao.versionField && version !== undefined && version !== null) extSql = appendSql.push("`" + dao.versionField + "` > " + u.int(version))
        if (dao.validField) extSql = appendSql.push("`" + dao.versionField + "` = 1")
        if (appendSql.length) extSql = extSql.replace(/where\s+/i, "WHERE " + appendSql.join(' AND '))
        if (!sqlArgs) sqlArgs = []
    }
    if (order) extSql += ' ORDER BY' + order
    return [extSql, sqlArgs]
}

const defaultPageSize = 20
let _db = db
let _redis = redis
let _userKey = 'id'

/**
 * 数据访问对象
 */
export default class {
    static config({ db, redis, userKey, idMaker }) {
        if (db) _db = db
        if (redis) _redis = redis
        if (userKey) _userKey = userKey
    }

    /**
     * 创建数据对象
     * @param {string} table 
     * @param {Object} info 
     * @param {string} info.pk 主键
     * @param {string} info.listFields 列表字段
     * @param {string} info.searchFields 允许搜索的字段
     * @param {string} info.getFields 详情字段
     * @param {string} info.updateFields 默认更新数据字段
     * @param {string} info.ignoreFields 对外忽略的字段
     * @param {string} info.makeID 生成主键的方法（如果设置，新增数据时使用该方法创建主键）
     * @param {Object} info.ownerField 所属人字段(如果设置，新增数据时使用该字段创建所属人，查询和编辑时会自动限制所属人)
     * @param {Object} info.versionField 版本字段(如果设置，更新和删除数据时使用该字段创建版本，获取列表时会根据版本号进行增量返回数据)
     * @param {Object} info.addUserField 添加人字段(如果设置，新增数据时使用该字段创建添加人)
     * @param {Object} info.addTimeField 添加时间字段(如果设置，新增数据时使用该字段创建添加时间)
     * @param {Object} info.updateUserField 更新人字段(如果设置，更新数据时使用该字段创建更新人)
     * @param {Object} info.updateTimeField 更新时间字段(如果设置，更新数据时使用该字段创建更新时间)
     * @param {Object} validField 有效字段(如果设置，删除数据时会将该字段设置为0，而不是真正删除数据，操作数据时会自动忽略无效数据)
     * @param {Object} info.conn 数据库连接
     */
    constructor(table, info) {
        this.table = table
        this.pk = (info.pk || 'id').split(', ')
        this.listFields = (info.listFields || info.getFields).split(', ')
        this.searchFields = (info.searchFields || info.updateFields).split(', ')
        this.getFields = info.getFields.split(', ')
        this.updateFields = info.updateFields.split(', ')
        this.ignoreFields = info.ignoreFields.split(', ')
        this.ownerField = info.ownerField || ''
        this.versionField = info.versionField || ''
        this.validField = info.validField || ''
        this.addUserField = info.addUserField || ''
        this.addTimeField = info.addTimeField || ''
        this.updateUserField = info.updateUserField || ''
        this.updateTimeField = info.updateTimeField || ''
        this.makeID = info.makeID
        this.db = info.db || _db
        this.redis = info.redis || _redis
        this.userKey = info.userKey || _userKey
    }

    /**
     * 获取列表
     * @param {Object} param - 查询参数对象
     * @param {number} param.page 页码（如果数据没有使用版本管理，进行分页）
     * @param {number} param.pageSize 每页大小
     * @param {string} param.fields 获取指定字段（默认使用this.listFields）
     * @param {Object} param.query 查询条件（如果包含%则使用LIKE，如果是数组使用IN）
     * @param {string} param.order 排序字段
     * @param {string} param.extSql 扩展查询条件
     * @param {Array} param.sqlArgs 扩展查询条件参数
     * @param {string} param.owner 所属人
     * @param {number} param.version 版本号（如果指定，根据版本号获取增量数据）
     * @returns {Object} - 返回查询结果，包含列表、最大版本号、总行数、总页数和每页大小
     */
    list({ args, session, page, pageSize, fields, query, order, extSql, sqlArgs, owner, version, isAdmin }) {
        try {
            if (args && !page) page = args.page
            if (args && !pageSize) pageSize = args.pageSize
            if (args && !query) query = args.query
            if (args && !order) order = args.order
            if (args && !version) version = args.version
            if (session && session.get && (!isAdmin || !owner)) owner = session.get(this.userKey)
            if (!page || page < 1) page = 1
            if (pageSize === undefined || pageSize === null) pageSize = defaultPageSize
            if (args && query && !isAdmin) query = u.getByKey(query, this.searchFields)
            let whereInfo = makeWhere(this, isAdmin, query, order, extSql, sqlArgs, owner, version)
            extSql = whereInfo[0]
            sqlArgs = whereInfo[1]
            let rowCount = 0
            let pageCount = 0
            let limit = ''
            let useVersion = this.versionField && version !== undefined && version !== null
            if (!useVersion) {
                if (pageSize > 0) {
                    rowCount = this.db.query(`SELECT COUNT(*) FROM ${this.table} ${extSql}`, ...sqlArgs).result
                    let rowStart = (page - 1) * pageSize
                    if (rowStart < 0) rowStart = 0
                    limit = ` LIMIT ${rowStart}, ${page * pageSize}`
                    pageCount = Math.ceil(rowCount / pageSize)
                }
            }
            let list = this.db.query(`SELECT ${fields || makeSelectByFields(this, isAdmin, this.listFields)} FROM ${this.table} ${extSql}${limit}`, ...sqlArgs).result
            if (useVersion) {
                let maxVersion = version || 0
                if (list.length > 0) maxVersion = list[list.length - 1][this.versionField]
                return { list, maxVersion }
            } else {
                return { list, rowCount, pageCount, pageSize }
            }
        } catch (ex) {
            log.error(ex)
            return { list: [], rowCount: 0, pageCount: 0, pageSize }
        }
    }

    /**
     * 获取单条数据
     * @param {Object} param - 查询参数对象
     * @param {string} [param.fields] - 指定要查询的字段，默认为this.getFields
     * @param {Object} [param.query] - 查询条件
     * @param {string} [param.extSql] - 扩展查询条件
     * @param {Array} [param.sqlArgs] - 扩展查询条件参数
     * @param {string} [param.owner] - 所属人
     * @returns {Object} - 返回查询结果，包含单条数据
     */
    get({ args, session, fields, query, extSql, sqlArgs, owner, isAdmin, fromAPI }) {
        try {
            if (args && !query) query = args.query
            if (args && !query) {
                if (this.pk.length) query = u.getByKey(args, this.pk)
                if (!this.pk.length || !query[this.pk[0]]) query = u.getByKey(args, this.getFields)
            }
            if (session && session.get && (!isAdmin || !owner)) owner = session.get(this.userKey)
            let whereInfo = makeWhere(this, isAdmin, query, '', extSql, sqlArgs, owner)
            extSql = whereInfo[0]
            sqlArgs = whereInfo[1]
            if (!fields) {
                if (fromAPI && this.ignoreFields.length) {
                    // 对外输出时排出忽略的字段
                    let getFields = []
                    for (let f of this.getFields) {
                        if (this.ignoreFields.indexOf(f) < 0) getFields.push(f)
                    }
                    fields = makeSelectByFields(this, isAdmin, getFields)
                } else {
                    fields = makeSelectByFields(this, isAdmin, this.getFields)
                }
            }
            return this.db.query1(`SELECT ${fields} FROM ${this.table} ${extSql} LIMIT 1`, ...sqlArgs).result
        } catch (ex) {
            log.error(ex)
            return {}
        }
    }

    /**
     * 获取新的版本号
     * @returns {number} - 版本号+1返回
     */
    fetchVersion() {
        if (!this.versionField) return 0
        let version = u.int(this.redis.incr('_DATA_VERSION_' + this.table))
        if (version > 1) {
            // 设置使用中的标记（确保数据更新完才被获取）
            this.redis.setEX('_DATA_VERSION_DOING_' + this.table + '_' + version, 10, true)
            return version
        }

        // 从数据库读取版本号进行初始化
        version = u.int(this.db.query11("SELECT MAX(`version`) FROM " + this.table).result) + 1
        this.redis.mSet('_DATA_VERSION_' + this.table, version, '_DATA_MAX_VERSION_' + this.table, version) // 初始化_DATA_MAX_VERSION_，用于版本在更新时存储最大可用版本号
        this.redis.setEX('_DATA_VERSION_DOING_' + this.table + '_' + version, 10, true)
        return version
    }

    /**
     * 提交版本号
     * @param {number} version - 版本号
     */
    commitVersion(version) {
        if (!this.versionField) return 0
        // 先存储当前版本完成标记，然后检查所有新版本是否完成以设置MAX_VERSION
        this.redis.del('_DATA_VERSION_DOING_' + this.table + '_' + version)
        let seqVersion = u.int(this.redis.get('_DATA_VERSION_' + this.table))
        let currentMaxVersion = u.int(this.redis.get('_DATA_MAX_VERSION_' + this.table))
        for (let i = currentMaxVersion; i <= seqVersion; i++) {
            if (this.redis.exists('_DATA_VERSION_DOING_' + this.table + '_' + version)) {
                // 遇到仍在处理的版本，跳过更新MAX_VERSION，确保用户获取数据是有序的
                break
            } else {
                // 更新MAX_VERSION，用户可以使用该版本
                this.redis.set('_DATA_MAX_VERSION_{{.TableName}}', i)
            }
        }
    }

    /**
     * 更新数据
     * @param {Object} param - 编辑参数对象
     * @param {Object} param.data - 数据对象
     * @param {string} [param.extSql] - 扩展查询条件
     * @param {Array} [param.sqlArgs] - 扩展查询条件参数
     * @param {string} [param.owner] - 所属人
     * @returns {number} - 返回受影响的行数（如果是新增数据返回自动生成的ID）
     */
    update({ args, session, data, extSql, sqlArgs, owner, isAdmin }) {
        try {
            if (args && !data) data = args.data
            if (args && !data) {
                data = u.getByKey(args, this.updateFields)
            } else {
                if (!isAdmin) data = u.getByKey(data, this.updateFields)
            }
            if (session && session.get && (!isAdmin || !owner)) owner = session.get(this.userKey)
            let version = 0
            let now = ''
            if (this.updateTimeField || this.addTimeField) now = u.toDatetime()
            if (this.updateTimeField) data[this.updateTimeField] = now
            if (this.updateUserField && owner) data[this.updateUserField] = owner
            if (this.versionField) {
                version = this.redis.incr(this.table + '_VERSION')
                data[this.versionField] = version
            }
            if (this.pk.length > 0 && !data[this.pk[0]]) {
                if (this.makeID) data[this.pk[0]] = this.makeID()
                if (this.addTimeField) data[this.addTimeField] = now
                if (this.addUserField && owner) data[this.addUserField] = owner
                if (this.ownerField && owner) data[this.ownerField] = owner
                let r = this.db.insert(this.table, data)
                if (this.versionField) this.commitVersion(version)
                return r.changes ? (this.makeID ? data[this.pk[0]] : r.id) : 0
            } else {
                let whereInfo = makeWhere(this, isAdmin, u.getByKey(data, this.pk), extSql, sqlArgs, owner)
                extSql = whereInfo[0]
                sqlArgs = whereInfo[1]
                extSql = extSql.replace(/where\s+/i, '')
                for (let k of this.pk) delete data[k]
                let r = this.db.update(this.table, data, extSql, ...sqlArgs).changes
                if (this.versionField) this.commitVersion(version)
                return r.changes
            }
        } catch (ex) {
            log.error(ex)
            return 0
        }
    }

    /**
     * 删除数据
     * @param {Object} param - 删除参数对象
     * @param {Object} param.query - 查询条件
     * @param {string} [param.extSql] - 扩展查询条件
     * @param {Array} [param.sqlArgs] - 扩展查询条件参数
     * @param {string} [param.owner] - 所属人
     * @returns {number} - 返回受影响的行数
     */
    remove({ args, session, query, extSql, sqlArgs, owner, isAdmin }) {
        try {
            if (args && !query) query = args.query
            if (args && !query) query = u.getByKey(args, this.pk)
            if (session && session.get && (!isAdmin || !owner)) owner = session.get(this.userKey)
            let whereInfo = makeWhere(this, isAdmin, query, '', extSql, sqlArgs, owner)
            extSql = whereInfo[0]
            sqlArgs = whereInfo[1]
            if (this.validField) {
                // 逻辑删除
                let data = { [this.validField]: 0 }
                let version = 0
                if (this.versionField) {
                    version = this.redis.incr(this.table + '_VERSION')
                    data[this.versionField] = version
                }
                let now = ''
                if (this.updateTimeField || this.addTimeField) now = u.toDatetime()
                if (this.updateTimeField) data[this.updateTimeField] = now
                if (this.updateUserField && owner) data[this.updateUserField] = owner
                extSql = extSql.replace(/where\s+/i, '')
                let r = this.db.update(this.table, data, extSql, ...sqlArgs).changes
                if (this.versionField) this.commitVersion(version)
                return r.changes
            } else {
                // 物理删除（先备份数据）
                let list = this.db.query(`SELECT * FROM ${this.table} ${extSql}`, ...sqlArgs).result
                let time = u.toDatetime()
                for (let item of list) {
                    let owner = this.ownerField ? (item[this.ownerField] || '') : ''
                    let pkValues = []
                    for (let pk of this.pk) pkValues.push(item[pk] || '')
                    let id = pkValues.join(',')
                    this.db.insert('_deleted', { table: this.table, id, owner, time, data: u.json(item) })
                }
                extSql = extSql.replace(/where\s+/i, '')
                return this.db.delete(this.table, extSql, ...sqlArgs).changes
            }
        } catch (ex) {
            log.error(ex)
            return 0
        }
    }

    /**
     * 复制一个使用事务的实例
     * @param {Object} tx - 事务对象
     * @returns {Dao} - 返回一个新的Dao对象，使用相同的数据库连接和配置
     */
    tx(tx) {
        let obj = u.copy(this)
        obj.conn = tx
        return obj
    }
}
